// Code to generate fragment programs for performing image convolutions using Cg

#include <stdio.h>
#include <glh/glh_extensions.h>
#include <glh/glh_array.h>
#include <shared/data_path.h>
#include "convolve.h"

extern CGcontext cg_context;

static CGprofile fragment_profile;
#define MAX_CG_ARGS 10
static const char *args[MAX_CG_ARGS]; // Cg compilation arguments
static data_path media;

// evaluate filter function over 1D domain
float *evaluateFilter1D(NVIFilter2D &filter, int radius, int &nsamples, bool normalize)
{
    nsamples = radius*2+1;
    float *weights = new float [nsamples];
 
    float total = 0.0f;
    for(int i=0; i<nsamples; i++) {
        float u = (float) i - radius;
        weights[i] = filter(u, 0.0f);
        total += weights[i];
    }

    // normalize weights
    if (normalize) {
        for(int i=0; i<nsamples; i++) {
            weights[i] /= total;
            //printf("%d: %f\n", i, weights[i]);
        }
    }

    return weights;
}

// evaluate filter function over 2D domain, returns an image
Image2D *evaluateFilter2D(NVIFilter2D &filter, int radius_x, int radius_y, bool normalize)
{
    int width = radius_x*2+1;
    int height = radius_y*2+1;
    Image2D *img = new Image2D(width, height);

    float total = 0.0f;
    for(int y=0; y<height; y++) {
        for(int x=0; x<width; x++) {
            float u = (float) x - radius_x;
            float v = (float) y - radius_y;
            (*img)(x, y) = filter(u, v);
            total += (*img)(x, y);
        }
    }

    if (normalize) {
        // normalize weights
        for(int y=0; y<height; y++) {
            for(int x=0; x<width; x++) {
                (*img)(x, y) /= total;
            }
        }
    }

    return img;
}

// generate a texture atlas containing a range of different kernel images
Image2D *evaluateFilterAtlas2D(NVIFilter2D &filter, int nkernels, int max_radius, float min_r, float max_r)
{
    int width = max_radius*2+1;
    int height = max_radius*2+1;
    Image2D *atlas_img = new Image2D(nkernels*width, height);

    for(int i=0; i<nkernels; i++) {
        float r = min_r + (max_r - min_r) * (i / (float)(nkernels-1));
        filter.width(r*2);
        filter.height(r*2);
        Image2D *kernel_img = evaluateFilter2D(filter, max_radius, max_radius, true);
        // copy into atlas image
        atlas_img->copy(*kernel_img, i*width, 0, kernel_img->get_width(), kernel_img->get_height());
        delete kernel_img;
    }

    return atlas_img;
}

// create a luminance texture from a 2D image
GLuint createTextureFromImage2D(Image2D &img, GLenum target, GLenum internalFormat)
{
    GLuint tex;
    glGenTextures(1, &tex);
    glBindTexture(target, tex);
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glTexImage2D(target, 0, internalFormat, img.get_width(), img.get_height(), 0,
                 GL_LUMINANCE, GL_FLOAT, (const void *)img.get_pointer());
    return tex;
}

// create a 1D texture from an array of floats
GLuint createTextureFromImage1D(float *data, int width, GLenum target, GLenum internalFormat)
{
    GLuint tex;
    glGenTextures(1, &tex);
    glBindTexture(target, tex);
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glTexImage1D(target, 0, internalFormat, width, 0,
                 GL_LUMINANCE, GL_FLOAT, data);
    return tex;
}

// initialize sample offsets for 1D kernel
// texel_size is value to multiply integer texture coordinates by - (1.0/width, 1.0/height) for 2D textures, (1, 1) for texture rectangles
vec2f *generateOffsets1D(int nsamples, int kernel_offset, bool vertical, vec2f texel_size)
{
    vec2f *offsets = new vec2f [nsamples];
    for(int i=0; i<nsamples; i++) {
        if (vertical) {
            offsets[i] = vec2f(0.0f, (float) i + kernel_offset) * texel_size;
        } else {
            offsets[i] = vec2f((float) i + kernel_offset, 0.0f) * texel_size;
        }
    }
    return offsets;
}

// generate weights and offsets for filter, taking advantage of linear texture filtering
// see GPU Gems 2, chapter 20
vec2f *generateOffsets1DLinear(float *weights, float **new_weights, int &nsamples, int kernel_offset, bool vertical, vec2f texel_size)
{
    // convert to smaller number of bilinear samples
    int new_nsamples = (nsamples+1) / 2;
    *new_weights = new float [new_nsamples];
    vec2f *new_offsets = new vec2f [new_nsamples];

    for(int i=0; i<new_nsamples; i++) {
        float a = weights[i*2];
        float b = (i*2+1 < nsamples-1) ? weights[i*2+1] : 0.0f;
        (*new_weights)[i] = a + b;
        float offset = b / (a + b);
//        assert(offset >= 0.0 && offset <= 1.0);
        if (vertical) {
            new_offsets[i] = vec2f(0.0f, (i*2 + kernel_offset) + offset) * texel_size;
        } else {
            new_offsets[i] = vec2f((i*2 + kernel_offset) + offset, 0.0f) * texel_size;
        }
//        printf("%d: %f, (%f, %f)\n", i, (*new_weights)[i], new_offsets[i][0], new_offsets[i][1]);
    }

    nsamples = new_nsamples;
    return new_offsets;
}

vec2f *generateOffsets2D(int width, int height, int x_offset, int y_offset, vec2f texel_size)
{
    vec2f *offsets = new vec2f [width * height];
    for(int y=0; y<height; y++) {
        for(int x=0; x<width; x++) {
            offsets[y*width+x] = vec2f(x + x_offset, y + y_offset) * texel_size;
        }
    }
    return offsets;
}

void init_convolve(bool use_half, int components, bool use_tex2D, bool use_looping)
{
    fragment_profile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
    cgGLSetOptimalOptions(fragment_profile);
    cgSetAutoCompile(cg_context, CG_COMPILE_MANUAL);

    // set compile options
    int arg = 0;
    char *components_str = new char[20];
    sprintf(components_str, "-DCOMPONENTS=%d", components);
    args[arg++] = components_str;

    if (use_half) {
        args[arg++] = strdup("-DUSE_HALF");
    }

    if (use_tex2D) {
        args[arg++] = strdup("-DsamplerIMG=sampler2D");
        args[arg++] = strdup("-DtexIMG=tex2D");
	} else {
        args[arg++] = strdup("-DsamplerIMG=samplerRECT");
        args[arg++] = strdup("-DtexIMG=texRECT");
	}

    if (fragment_profile == CG_PROFILE_FP40) {
        if (!use_looping) {
            // always unroll loops
            // note: have to use separate strings for each arg!
            args[arg++] = strdup("-unroll");
            args[arg++] = strdup("all");
        }

        // use branching for ifs
        args[arg++] = strdup("-ifcvt");
        args[arg++] = strdup("none");
    }

    args[arg++] = 0;

    media.path.clear();
    media.path.push_back(".");
    media.path.push_back("../../../MEDIA/programs/imgproc_convolve");
    media.path.push_back("../../../../MEDIA/programs/imgproc_convolve");
}

void destroy_convolve()
{
    for(int i=0; i<MAX_CG_ARGS; i++) {
        if (args[i]) {
            free((void *) args[i]);
            args[i] = 0;
        }
    }
}

// Use Cg to generate a fragment program for performing a convolution with given weights and offsets
CGprogram createConvolveProgram(int nsamples, float *weights, vec2f *offsets, bool use_tex2D)
{
    // load program
    CGprogram fprog;
    fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE,
                                    media.get_file("convolve.cg").c_str(),
                                    fragment_profile, "convolve", args);
    if (!fprog) {
        fprintf(stderr, "Error creating convolution fragment program.\n");
        return 0;
    }

    // set parameters
    CGparameter nsamples_param, weights_param, offsets_param;
    nsamples_param = cgGetNamedParameter(fprog, "nsamples");
    if (!nsamples_param) {
        return 0;
    }
    cgSetParameter1i(nsamples_param, nsamples);
    cgSetParameterVariability(nsamples_param, CG_LITERAL);

    weights_param = cgCreateParameterArray(cg_context, CG_HALF, nsamples);
    offsets_param = cgCreateParameterArray(cg_context, CG_FLOAT2, nsamples);
    for(int i=0; i<nsamples; i++) {
        cgSetParameter1f(cgGetArrayParameter(weights_param, i), weights[i]);
        cgSetParameter2f(cgGetArrayParameter(offsets_param, i), offsets[i][0], offsets[i][1]);
    }
    cgSetParameterVariability(weights_param, CG_LITERAL);
    cgSetParameterVariability(offsets_param, CG_LITERAL);

    cgConnectParameter(weights_param, cgGetNamedParameter(fprog, "weight"));
    cgConnectParameter(offsets_param, cgGetNamedParameter(fprog, "offset"));

    cgCompileProgram(fprog);
    cgGLLoadProgram(fprog);

#if 0
    const char *obj_code = cgGetProgramString(fprog, CG_COMPILED_PROGRAM);
    printf("%s\n", obj_code);
#endif

    return fprog;
}


// create fragment program for performing separable convolution
CGprogram createConvolutionProgramSeparable(NVIFilter2D &filter, int radius, bool useFiltering, bool vertical, bool use_tex2D, int image_w, int image_h)
{
    int nsamples;
    float *weights = evaluateFilter1D(filter, radius, nsamples, true);

    vec2f texel_size;
    if (use_tex2D) {
        texel_size = vec2f(1.0f / image_w, 1.0f / image_h);
    } else {
        texel_size = vec2f(1.0f, 1.0f);
    }

    vec2f *offsets;
    if (useFiltering) {
        float *new_weights, *temp;
        offsets = generateOffsets1DLinear(weights, &new_weights, nsamples, -radius, vertical, texel_size);
        temp = weights;
        weights = new_weights;
        delete [] temp;
    } else {
        offsets = generateOffsets1D(nsamples, -radius, vertical, texel_size);        
    }

    CGprogram fprog = createConvolveProgram(nsamples, weights, offsets, use_tex2D);
    delete [] weights;
    delete [] offsets;

    return fprog;
}

CGprogram createConvolutionProgram2D(NVIFilter2D &filter, int radius_x, int radius_y, bool use_tex2D, int image_w, int image_h)
{
    Image2D *kernel = evaluateFilter2D(filter, radius_x, radius_y, true);

    vec2f texel_size;
    if (use_tex2D) {
        texel_size = vec2f(1.0f / image_w, 1.0f / image_h);
    } else {
        texel_size = vec2f(1.0f, 1.0f);
    }
    vec2f *offsets = generateOffsets2D(kernel->get_width(), kernel->get_height(), -radius_x, -radius_y, texel_size);

    CGprogram fprog = createConvolveProgram(kernel->get_width() * kernel->get_height(), kernel->get_pointer(), offsets, use_tex2D);

    delete kernel;
    delete [] offsets;
    return fprog;
}


// create fragment program for performing convolution with kernel stored in texture
CGprogram createConvolveProgramTex(int kernel_width, int kernel_height, int xoffset, int yoffset, bool use_tex2D, int img_width, int img_height)
{
    // load program
    CGprogram fprog;
    fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("convolve.cg").c_str(),
                                    fragment_profile, "convolve_tex_2D", args);
    if (!fprog) {
        fprintf(stderr, "Error creating convolution fragment program.\n");
        return 0;
    }

    // set kernel width and height literal parameters
    CGparameter kernel_width_param = cgGetNamedParameter(fprog, "kernel_width");
    CGparameter kernel_height_param = cgGetNamedParameter(fprog, "kernel_height");
    CGparameter kernel_offset_param = cgGetNamedParameter(fprog, "kernel_offset");
    CGparameter texel_size_param = cgGetNamedParameter(fprog, "texel_size");
    cgSetParameterVariability(texel_size_param, CG_LITERAL);
    if (use_tex2D) {
        cgSetParameter2f(texel_size_param, 1.0f/img_width, 1.0f/img_height);
    } else {
        cgSetParameter2f(texel_size_param, 1.0f, 1.0f);
    }

    cgSetParameter1i(kernel_width_param, kernel_width);
    cgSetParameter1i(kernel_height_param, kernel_height);
    cgSetParameter2i(kernel_offset_param, xoffset, yoffset);

    cgSetParameterVariability(kernel_width_param, CG_LITERAL);
    cgSetParameterVariability(kernel_height_param, CG_LITERAL);
    cgSetParameterVariability(kernel_offset_param, CG_LITERAL);

    cgCompileProgram(fprog);
    cgGLLoadProgram(fprog);

    return fprog;
}

CGprogram createConvolveProgramTexSeparable(int kernel_width, int offset, bool use_tex2D, bool vertical, int img_width, int img_height)
{
    // load program
    CGprogram fprog;
    fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("convolve.cg").c_str(),
                                    fragment_profile, "convolve_tex_1D", args);
    if (!fprog) {
        fprintf(stderr, "Error creating convolution fragment program.\n");
        return 0;
    }

    // set kernel width and height literal parameters
    CGparameter kernel_width_param = cgGetNamedParameter(fprog, "kernel_width");
    CGparameter kernel_offset_param = cgGetNamedParameter(fprog, "kernel_offset");
    CGparameter texel_size_param = cgGetNamedParameter(fprog, "texel_size");
    cgSetParameterVariability(texel_size_param, CG_LITERAL);
    if (vertical) {
        if (use_tex2D) {
            cgSetParameter2f(texel_size_param, 0.0f, 1.0f/img_height);
        } else {
            cgSetParameter2f(texel_size_param, 0.0f, 1.0f);
        }
    } else {
        if (use_tex2D) {
            cgSetParameter2f(texel_size_param, 1.0f/img_width, 0.0f);
        } else {
            cgSetParameter2f(texel_size_param, 1.0f, 0.0f);
        }
    }
    cgSetParameter1i(kernel_width_param, kernel_width);
    cgSetParameter1f(kernel_offset_param, offset);

    cgSetParameterVariability(kernel_width_param, CG_LITERAL);
    cgSetParameterVariability(kernel_offset_param, CG_LITERAL);

    cgCompileProgram(fprog);
    cgGLLoadProgram(fprog);

    return fprog;
}


static CGparameter depth_scale_param = 0, depth_bias_param = 0;

// create fragment program for performing steerable convolution with kernel stored in texture
CGprogram createConvolveProgramTexSteerable(int kernel_width, int kernel_height, int xoffset, int yoffset, int nkernels, bool use_tex2D, int img_width, int img_height)
{
    // load program
    CGprogram fprog;
    fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, media.get_file("convolve.cg").c_str(),
                                    fragment_profile, "convolve_tex_steerable", args);
    if (!fprog) {
        fprintf(stderr, "Error creating convolution fragment program.\n");
        return 0;
    }

    // set kernel width and height literal parameters
    CGparameter kernel_width_param = cgGetNamedParameter(fprog, "kernel_width");
    CGparameter kernel_height_param = cgGetNamedParameter(fprog, "kernel_height");
    CGparameter kernel_offset_param = cgGetNamedParameter(fprog, "kernel_offset");
    CGparameter nkernels_param = cgGetNamedParameter(fprog, "nkernels");
    depth_scale_param = cgGetNamedParameter(fprog, "depth_scale");
    depth_bias_param = cgGetNamedParameter(fprog, "depth_bias");
    CGparameter texel_size_param = cgGetNamedParameter(fprog, "texel_size");
    cgSetParameterVariability(texel_size_param, CG_LITERAL);
    if (use_tex2D) {
        cgSetParameter2f(texel_size_param, 1.0f/img_width, 1.0f/img_height);
    } else {
        cgSetParameter2f(texel_size_param, 1.0f, 1.0f);
    }
    cgSetParameter1i(kernel_width_param, kernel_width);
    cgSetParameter1i(kernel_height_param, kernel_height);
    cgSetParameter2i(kernel_offset_param, xoffset, yoffset);
    cgSetParameter1i(nkernels_param, nkernels - 1);

    cgSetParameterVariability(kernel_width_param, CG_LITERAL);
    cgSetParameterVariability(kernel_height_param, CG_LITERAL);
    cgSetParameterVariability(kernel_offset_param, CG_LITERAL);
    cgSetParameterVariability(nkernels_param, CG_LITERAL);

    cgCompileProgram(fprog);
    cgGLLoadProgram(fprog);

    return fprog;
}

void setDepthScaleBias(float scale, float bias)
{
    if (depth_scale_param && depth_bias_param) {
        cgSetParameter1f(depth_scale_param, scale);
        cgSetParameter1f(depth_bias_param, bias);
    }
}
